/*
* Copyright(c) 2019 Netflix, Inc.
*
* This source code is subject to the terms of the BSD 2 Clause License and
* the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
* was not distributed with this source code in the LICENSE file, you can
* obtain it at https://www.aomedia.org/license/software-license. If the Alliance for Open
* Media Patent License 1.0 was not distributed with this source code in the
* PATENTS file, you can obtain it at https://www.aomedia.org/license/patent-license.
*/

#ifndef EbDecStruct_h
#define EbDecStruct_h

#include "EbPictureBufferDesc.h"
#include "EbBlockStructures.h"
#include "EbSvtAv1Dec.h"
#include "EbAv1Structs.h"

#ifdef __cplusplus
extern "C" {
#endif

/* TO enable some debug checks for coeff producer and consumer */
#define SVT_DEC_COEFF_DEBUG 0

#ifdef __cplusplus
}
#endif
#endif // EbDecStruct_h
